/* $Id: fdt.cpp,v 1.7 2004/12/15 17:19:31 roca Exp $ */
/*
 *  Copyright (c) 2003-2004 INRIA - All rights reserved
 *  main authors: Christoph Neumann - christoph.neumann@inrialpes.fr
 *		  Vincent Roca - vincent.roca@inrialpes.fr
 *		  Julien Laboure - julien.laboure@inrialpes.fr
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#include "flute.h"
XERCES_CPP_NAMESPACE_USE

 /*local variables*/
XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument* doc;
DOMImplementation *impl;
DOMBuilder* parser;


 /*** Begin of private functions ***/

int hasTOIinFDT(unsigned int toi){

	XMLCh tempStr[100];
	unsigned int i, nchilds;
	int hasTOI = 0;

	try
	{
	
	  	/* I get reference to the root element of the document */
 		DOMElement*   root = doc->getDocumentElement();
			
		XMLString::transcode("File", tempStr, 99);
		DOMNodeList*  node_list = root->getElementsByTagName(tempStr);
		
		nchilds = node_list->getLength();
		XMLString::transcode("TOI", tempStr, 99);
		for (i = 0; i < nchilds; i++) {
			DOMNode* node = node_list->item(i);
			DOMAttr*    pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
			const XMLCh*        toi_value = pAttr->getValue();
			if(toi==(unsigned int) XMLString::parseInt(toi_value))
			{
				hasTOI=1;
				break;	
			}
		}


        }
        catch (const DOMException& toCatch) {
            char* message = XMLString::transcode(toCatch.getMessage());
            EXIT(("Error during hasTOIinFDT(unsigned int toi):\n\t %s\n", message))
        }	

	
	return hasTOI;

}
 /*** End of private functions ***/


 /* Creates the FDT */
 void createFDT()
 {

	XMLCh tempStr[100];
	
#if 0
	time_t systime;
	char * str_time = (char *) malloc(10); //10 is maximum length of UL string
#endif

	try
	{
		XMLPlatformUtils::Initialize();
        }
        catch (const XMLException& toCatch) {
            char* message = XMLString::transcode(toCatch.getMessage());
            EXIT(("Error during initialization:\n\t %s\n", message))
        }	

	/* First I get a DOMImplementation reference */	
        XMLString::transcode("XML 1.0 XML 2.0 Range 2.0 Traversal 2.0", tempStr, 99);
        impl = DOMImplementationRegistry::getDOMImplementation(tempStr);
	
	/* Initialize the parser */	
	parser = ((DOMImplementationLS*)impl)->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);

	/* I create a new document with FDT as root element */
	XMLString::transcode("FDT", tempStr, 99);
        doc = impl->createDocument(0, tempStr, 0);

#if 0
	time(&systime);
	sprintf(str_time,"%lu",systime + 2208988800);
	setExpires(str_time);
	free(str_time);
#endif

	setExpires("4294967294");
	
 }
 
 
 
 void setExpires(char * ivalue)
 {
 
	XMLCh tempStr[100];
 
 
	try
	{
	  	/* I get reference to the root element of the document */
 		DOMElement*   root = doc->getDocumentElement();
 	
		XMLString::transcode("Expires", tempStr, 99);
       	 	DOMAttr*       pAttr = doc->createAttribute(tempStr);
       	 	root->setAttributeNode(pAttr);

		pAttr->setValue((const XMLCh *)ivalue);
        }
        catch (const DOMException& toCatch) {
            char* message = XMLString::transcode(toCatch.getMessage());
            EXIT(("Error during setExpires(char * ivalue):\n\t %s\n", message))
        }	

	
 }

char * getExpires()
 {	
 
	XMLCh tempStr[100];
	char * result;

	try
	{ 
  		/* I get reference to the root element of the document */
 		DOMElement*   root = doc->getDocumentElement();
 
		/*Get the expires attribute*/
 		XMLString::transcode("Expires", tempStr, 99);
		DOMAttr*      pAttr = root->getAttributeNode(tempStr);
	
		const XMLCh* 	      value = pAttr->getValue();
 	
		result = (char *) value;
	}
        catch (const XMLException& toCatch) {
            char* message = XMLString::transcode(toCatch.getMessage());
            EXIT(("Error during getExpires():\n\t %s\n", message))
        }
	
	
	return result;
 }

 /* Closes the FDT */
 void closeFDT()
 {	
 	/* I free the document structure */
	doc->release();
	
	XMLPlatformUtils::Terminate();
 }
 
 
 /* Adds elementsof the FDTinstance to the FDT */
void updateFDT(char* xmlfdtinstance, unsigned int length){
	
	XMLCh tempStr[100];
	
	/* I get reference to the root element of the document */
 	DOMElement*   root = doc->getDocumentElement();


	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument* fdtinstance;
	const DOMInputSource* domIS = new Wrapper4InputSource(new MemBufInputSource((const XMLByte *) xmlfdtinstance,length, tempStr));
	
	try {
            	fdtinstance = parser->parse(*domIS);
        }
        catch (const XMLException& toCatch) {
		char* message = XMLString::transcode(toCatch.getMessage());
	        EXIT(("Exception message is: %s\n", message))
	}
	
  	/* I get reference to the root element of the document */
 	DOMElement*   fdt_instance_root;
	if ((fdt_instance_root = fdtinstance->getDocumentElement()) == NULL)
	{
		EXIT(("Root element is null\n"))
	}

	/* Update Expires node */
 	XMLString::transcode("Expires", tempStr, 99);
	DOMAttr*      pAttr = fdt_instance_root->getAttributeNode(tempStr);
	const XMLCh* 	      value = pAttr->getValue();
	char * temp_char = XMLString::transcode(value);
	setExpires(temp_char);
	
	/* Go through the node of the instance and check if it is already included in the FDT */	
	DOMNodeList*  node_list = fdt_instance_root->getChildNodes();

	long i, nchilds;
	nchilds = node_list->getLength();
	XMLString::transcode("TOI", tempStr, 99);
	for (i = 0; i < nchilds; i++) {
		DOMNode* node = node_list->item(i);
		if (node->getNodeType() == DOMNode::ELEMENT_NODE)
		{
			pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
			value = pAttr->getValue();
			if(! hasTOIinFDT(XMLString::parseInt(value)) )
			{
				DOMNode* new_node = doc->importNode(node,true);
				root->appendChild(new_node);
			}
		}
	}
	
	fdtinstance->release();
} 

/* 
 * Adds element with TOI=@toi of the FDT to the FileList @filelist
 */

pFFile addTOItoFFile(pFFile filelist, unsigned int toi)
{
	XMLCh tempStr[100];
	DOMNode*   node = NULL;
	unsigned int i, nchilds;	

  	/* I get reference to the root element of the document */
 	DOMElement*   root = doc->getDocumentElement();
	
	XMLString::transcode("File", tempStr, 99);
	DOMNodeList*  node_list = root->getElementsByTagName(tempStr);
		
	nchilds = node_list->getLength();
	XMLString::transcode("TOI", tempStr, 99);
	for (i = 0; i < nchilds; i++) {
		DOMNode* temp_node = node_list->item(i);
		DOMAttr*    pAttr = ((DOMElement *) temp_node)->getAttributeNode(tempStr);
		const XMLCh*        toi_value = pAttr->getValue();
		if(toi==(unsigned int) XMLString::parseInt(toi_value))
		{
			node = temp_node;
			break;	
		}
	}
	
	
	
	if (node == NULL) {
		return filelist;
	}
	else
	{
				FFile NewFile;
				
				/* Content-Location */
				XMLString::transcode("Content-Location", tempStr, 99);
				DOMAttr*      pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
				const XMLCh* 	      value = pAttr->getValue();
				char * temp_char = XMLString::transcode(value);
				
				/* Remove string file:// */
				if (XMLString::startsWith(temp_char,"file://"))
				{
					int len = strlen(temp_char);
					strncpy(temp_char, temp_char + 7, len - 7);
					temp_char[len - 7] = '\0';
				}
				
				/* Detect possible ':' (like c:/) and replace them with '/' */
				char *p;
				if ((p = strchr(temp_char, ':')) != NULL) {
					*p = '/';
				}

				strncpy(NewFile.fullname,"./", 2); /* just to be sure it is copied in current directory */
				strncpy(NewFile.fullname+2, temp_char, MAX_PATH+MAX_FILENAME );
				NewFile.writeIt = CheckWriteContext(NewFile.fullname, overwrite);
				NewFile.received=0;
				NewFile.toi = toi;
				NewFile.integrity =0;
#if 0	/* MCL now opens and closes this file */
				if (NewFile.writeIt)
				{
#ifdef WIN32
					if ((NewFile.fd = open(	NewFile.fullname, O_WRONLY | O_CREAT | O_BINARY | O_TRUNC, _S_IWRITE) ) < 0)
#else
					if ((NewFile.fd = open(	NewFile.fullname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU)) < 0)
#endif
					{
						EXIT(("Error while opening file \"%s\"\n", NewFile.fullname))
					}
				}
#endif
				NewFile.next = NULL;
				
				/* Content-Length */
				NewFile.filesize=0;
				XMLString::transcode("Content-Length", tempStr, 99);
				if ((pAttr = ((DOMElement *) node)->getAttributeNode(tempStr))!=NULL)
				{
					value = pAttr->getValue();
					NewFile.filesize = (unsigned long) XMLString::parseInt(value);
				}
#ifdef WIN32
				ZeroMemory(NewFile.md5sum, MD5BASE64_LENGTH);
#else
				/* Content-MD5 */
				XMLString::transcode("Content-MD5", tempStr, 99);
				memset(NewFile.md5sum, 0, MD5BASE64_LENGTH);
				if ((pAttr = ((DOMElement *) node)->getAttributeNode(tempStr))!=NULL)
				{
					value = pAttr->getValue();
					char * temp_char = XMLString::transcode(value);
					strncpy((char *)NewFile.md5sum ,temp_char,MD5BASE64_LENGTH);
				}				
#endif	
				FFileInsert(&filelist, NewFile);

				
	}	
	
	return filelist;
	
}


/* 
 * Adds elements of the FDT to the FileList 
 * ONLY in !interactive mode
 */
  
pFFile updateFFile(pFFile filelist)
{
	
	XMLCh tempStr[100];
	
	/* I get reference to the root element of the document */
 	DOMElement*   root = doc->getDocumentElement();

	/* Go through the nodelist*/	
	DOMNodeList*  node_list = root->getChildNodes();

	long i, nchilds;
	nchilds = node_list->getLength();
	XMLString::transcode("TOI", tempStr, 99);
	for (i = 0; i < nchilds; i++) {
		DOMNode* node = node_list->item(i);
		if (node->getNodeType() == DOMNode::ELEMENT_NODE)
		{
			DOMAttr*    pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
			const XMLCh* 	   value = pAttr->getValue();
			if(((FFileFindTOI(XMLString::parseInt(value), filelist))==NULL))
			{

				filelist=addTOItoFFile(filelist, XMLString::parseInt(value));
				{
				int mcl_option = XMLString::parseInt(value);
				if (mcl_ctl(id, MCL_OPT_FLUTE_DELIVER_THIS_ADU_TO_APPLI, (void*)&mcl_option, sizeof(mcl_option)))
					EXIT(("mcl_ctl: MCL_OPT_FLUTE_DELIVER_THIS_ADU_TO_APPLI failed\n"))
				}
							
			}
		}
	}

	return filelist;

} 


/**
 * Displays the FDT on stdout.
 * @param highlight	index of line highlighted in the DOM node list.
 * @param filelist	pointer to the expected files.
 * @return		TOI of the highligted line.
 */
unsigned int displayFDT(unsigned int highlight, pFFile filelist)
{
	XMLCh		tempStr[100];
	unsigned int	i; /* current index while crossing the DOM node list. */
	unsigned int	j; /* current index for elements of type FILE in the DOM node list. */
	unsigned int	nchilds;
	unsigned int	returntoi = 0;	/* TOI of the highligted line */
#define WITH_FILE_RX_INFO
#ifdef WITH_FILE_RX_INFO
	mcl_adu_rx_info_t rx_info; /* struct containing ADU reception info */
#endif

#ifndef WIN32
	/*clear screen*/
	if (verbose==0)	printf("\e[H\e[J");
#endif
	/* I get reference to the root element of the document */
 	DOMElement*   root = doc->getDocumentElement();
	/*
	 * Go through the node of the instance and check if it is already
	 * included in the FDT
	 */	
	DOMNodeList*  node_list = root->getChildNodes();
	/*
	 * Go through the node of the instance and check if it is already
	 * included in the FDT
	 */
	if ((nchilds = node_list->getLength())==0) {	
		printf("FDT is empty\n");
		goto exit;
	}

	printf("TOI\t");
	if (filelist != NULL) {
		printf("Select\tRcvd\tLength\tMD5sum\t");
	}
	printf("Content-Location\n\n");
	/*
	 * Now go through all entries in the DOM node list...
	 */
	for (i = 0, j = 0; i < nchilds; i++) {
		DOMNode* node = node_list->item(i);
		if (node->getNodeType() == DOMNode::ELEMENT_NODE) {
			/* this is an entry of interest to us */
			XMLString::transcode("Content-Location", tempStr, 99);
			DOMAttr*	pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
			const XMLCh	*URIvalue = pAttr->getValue();

			XMLString::transcode("TOI", tempStr, 99);
			pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
			const XMLCh*	toi_value = pAttr->getValue();
			const unsigned int toi = XMLString::parseInt(toi_value);
#ifndef WIN32
			if (j == highlight)
				printf("\e[7m");	
#endif
			printf(" %i \t", toi);
			if (filelist != NULL) {
				pFFile	afile;	/* file associated to TOI */
				if ((afile = FFileFindTOI(toi, filelist)) != NULL) {
					/* yes, this file is expected! */
					printf(" X \t");
#define WITH_FILE_RX_INFO
#ifdef WITH_FILE_RX_INFO
					rx_info.toi = toi;
					if (mcl_ctl(id, MCL_OPT_GET_ADU_RX_INFO, (void*)&rx_info, sizeof(rx_info)) == 0) {
						double	val;	/* percentage */
						if (rx_info.len > 0)
							val = (double)rx_info.recvd_src_data / (double)rx_info.len * 100.0;
						else
							val = -1;
						printf(" %.1f\% \t", val);
					} else {
						printf("\t");
					}
#else /* WITH_FILE_RX_INFO */
					if (afile->received) printf(" X \t");
					else printf("\t");
#endif /* WITH_FILE_RX_INFO */

					if (afile->filesize == 0) printf("?\t");
					else if(afile->filesize < 0)  printf("NOK\t");
					else  printf("%ld\t", afile->filesize);

					if (afile->integrity == 1) printf(" OK \t");
					else if (afile->integrity == -1) printf(" NOK \t");
					else printf(" ? \t");
				}
			}
			char * temp_char = XMLString::transcode(URIvalue);
			printf("%s ",temp_char);

			printf("\n");
#ifndef WIN32
			if (j == highlight) {
				printf("\e[0m");
				returntoi = toi;
			}
#endif
			j++;			
		}
	}

	/*Get the expires value*/
	printf("\nExpires: %s (max: %lu)\n",  getExpires(), ULONG_MAX);
		
exit:
	return returntoi;	       
}


/* 
 * Create a new File Element, that may be added to the FDT
 */

DOMElement* createNewFile()
{

	XMLCh tempStr[100];

	XMLString::transcode("File", tempStr, 99);
	DOMElement*     el = doc->createElement(tempStr);
	
	return el;
}

/* 
 * Set attribute of a File Element
 */

DOMElement* setFileAttribute(DOMElement *el, char* iname, char* ivalue)
{

	XMLCh tempStr1[100];
	XMLCh tempStr2[100];
	
	XMLString::transcode(ivalue, tempStr2, 99);
	XMLString::transcode(iname, tempStr1, 99);

	/* fix the URI if it is the Content-Location attribute */
	if (strcmp(iname,"Content-Location")==0)
	{
		XMLCh tempStr3[100];
		XMLString::transcode(ivalue, tempStr3, 99);
		XMLString::fixURI(tempStr3, tempStr2);
	}

	el->setAttribute (tempStr1, tempStr2);

	return el;
}


/* 
 * Append a File Element to the FDT
 */

void appendFile(DOMElement *el)
{

	/* I get reference to the root element of the document */
 	DOMElement*   root = doc->getDocumentElement();

	/* append the node */
 	root->appendChild (el);

}


XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument* createNewFDTinstance()
{ 	


	XMLCh tempStr1[100];
	XMLCh tempStr2[100];	

	/* I create a new document with FDT as root element */
	XMLString::transcode("FDT-Instance", tempStr1, 99);
    XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument*    fdtinstance = impl->createDocument(0, tempStr1, 0);
	
	/* I get reference to the root element of the document */
 	DOMElement*   root = fdtinstance->getDocumentElement();

	/* I add the Expires attribute to the element  */	
	XMLString::transcode("Expires", tempStr1, 99);
	XMLString::transcode(getExpires(), tempStr2, 99);
	root->setAttribute (tempStr1, tempStr2);
		
	return fdtinstance;
}


XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument* AddFileToFDTinstance(XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument* fdtinstance ,unsigned int itoi)
{

	XMLCh tempStr[100];
	
	/* I get reference to the root element of the document */
 	DOMElement*   root = doc->getDocumentElement();

	/* Go through the nodelist*/	
	DOMNodeList*  node_list = root->getChildNodes();

	long i, nchilds;
	nchilds = node_list->getLength();
	XMLString::transcode("TOI", tempStr, 99);
	for (i = 0; i < nchilds; i++) {
		DOMNode* node = node_list->item(i);
		if (node->getNodeType() == DOMNode::ELEMENT_NODE)
		{
			DOMAttr*    pAttr = ((DOMElement *) node)->getAttributeNode(tempStr);
			const XMLCh*        toi_value = pAttr->getValue();
			if(itoi==(unsigned int) XMLString::parseInt(toi_value))
			{
				/* I get reference to the root element of the document */
			 	DOMElement*   root_instance = fdtinstance->getDocumentElement();
				DOMNode* new_node = fdtinstance->importNode(node,true);
				root_instance->appendChild(new_node);
				
			}

		}
	}
	
	return fdtinstance;
	
}


unsigned int getFinalFDTInstance(XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument* fdtinstance, char ** buffer)
{	
	XMLCh tempStr2[100];
	unsigned int len = 0;

	DOMWriter* theSerializer = ((DOMImplementationLS*)impl)->createDOMWriter();

	XMLString::transcode("UTF-8", tempStr2, 99);
	theSerializer->setEncoding(tempStr2);
	
	MemBufFormatTarget *  target = new MemBufFormatTarget(FLUTE_MAX_FDT_SIZE);
	
	theSerializer->writeNode(target,*fdtinstance);
	
	*buffer = (char*) target->getRawBuffer();
	
	len = target->getLen();
	
	theSerializer->release();	
	
	return len;
}


